from pyrevm import *


class Revm:
    _instance = None

    def __new__(cls, fork_url="", block_number=0, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(Revm, cls).__new__(cls)
            cls._instance.revm = EVM(
                fork_url=fork_url,
                tracing=False,
                fork_block_number=block_number,
            )
        return cls._instance

    def __init__(self, fork_url="", block_number=0):
        pass

    def call_raw(self, *args, **kwargs):
        return self.revm.call_raw(*args, **kwargs)

    def call_raw_committing(self, *args, **kwargs):
        return self.revm.call_raw_committing(*args, **kwargs)

    def get_balance(self, address):
        return self.revm.get_balance(address)



import json

from .revm import Revm
from .models import ABIFunction, ContractABI


ZERO_ADDRESS = "0x0000000000000000000000000000000000000000"

class Contract:
    def __init__(
        self,
        address: str,
        abi: dict = None,
        abi_file_path: str = None,
        caller: str = ZERO_ADDRESS,
        fork_url="",
        block_number=0,
    ):
        self.address = address
        self.caller = caller

        self.abi = self._load_abi(abi, abi_file_path)
        self.revm = Revm(fork_url=fork_url, block_number=block_number)

    def _load_abi(self, abi: dict = None, file_path: dict = None) -> ContractABI:
        if not abi and not file_path:
            raise ValueError("Either abi or abi_file_path must be provided")

        if file_path:
            with open(file_path, "r") as file:
                abi = json.load(file)

        functions = []
        for entry in abi:
            if entry["type"] == "function":
                name = entry["name"]
                inputs = [input["type"] for input in entry["inputs"]]
                outputs = [output["type"] for output in entry["outputs"]]
                constant = entry["stateMutability"] in ("view", "pure")
                payable = entry["stateMutability"] == "payable"
                functions.append(
                    ABIFunction(
                        name=name,
                        inputs=inputs,
                        outputs=outputs,
                        constant=constant,
                        payable=payable,
                    )
                )

        return ContractABI(functions)
    
    def __getattr__(self, attribute):
        for func in self.abi.functions:
            if func.name == attribute or func.selector == attribute:
                return lambda *args, **kwargs: self.call_function(
                    func, args, kwargs
                )
        raise AttributeError(
            f"No function named or matching selector {attribute} in"
            " contract ABI"
        )

    def call_function(self, func: ABIFunction, args: tuple, kwargs: dict = {}):
        calldata = func.encode_inputs(args)
        value = kwargs.get("value", 0)
        print(calldata.hex())
        if func.constant:
            raw_output = self.revm.call_raw(
                caller=self.caller, to=self.address, data=calldata
            )
            if isinstance(raw_output, str):
                raw_output = bytes.fromhex(
                    raw_output[2:]
                    if raw_output.startswith("0x")
                    else raw_output
                )
            return func.decode_outputs(raw_output)
        else:
            if not func.payable and value > 0:
                raise ValueError("Cannot send value to a non-payable function")

            if self.caller == ZERO_ADDRESS:
                raise ValueError("Cannot call a non-constant function without a caller")

            raw_output = self.revm.call_raw_committing(
                caller=self.caller,
                to=self.address,
                data=calldata,
                value=value,
            )
            if func.outputs:
                if isinstance(raw_output, str):
                    raw_output = bytes.fromhex(
                        raw_output[2:]
                        if raw_output.startswith("0x")
                        else raw_output
                    )
                return func.decode_outputs(raw_output)
            else:
                return raw_output

    def balance(self):
        return self.revm.get_balance(self.address)



from typing import List, Optional, Any
from dataclasses import dataclass

from eth_abi import encode, decode
import sha3


@dataclass
class ABIFunction:
    name: Optional[str] = None
    inputs: Optional[List[str]] = None
    possible_inputs: Optional[List[str]] = None
    num_args: Optional[int] = None
    outputs: List[str] = None
    constant: bool = False
    payable: bool = False
    selector: Optional[str] = None
    
    @staticmethod
    def func_selector(sig: str) -> bytes:
        k = sha3.keccak_256()
        k.update(sig.encode('utf-8'))
        return k.digest()[:4]

    def get_selector(self) -> bytes:
        if self.selector:
            return bytes.fromhex(
                self.selector[2:]
                if self.selector.startswith("0x")
                else self.selector
            )
        elif self.name:
            return self.func_selector(self.get_signature())
        else:
            raise ValueError(
                "Cannot compute selector without a name or selector"
            )

    def get_signature(self) -> str:
        if self.name is None:
            raise ValueError("Cannot compute signature without a name")

        input_types = ""
        if self.inputs:
            input_types = ",".join(self.inputs)

        return f"{self.name}({input_types})"

    def encode_inputs(self, values: List[Any]) -> bytes:
        selector = self.get_selector()
        if self.inputs == []:
            return selector
        encoded_inputs = encode(self.inputs, values)
        return selector + encoded_inputs

    def decode_outputs(self, output_data: bytes) -> Any:
        if not output_data:
            return None

        if not self.outputs:
            return []

        try:
            decoded = decode(self.outputs, output_data)
            if len(decoded) == 1:
                return decoded[0]
            return decoded
        except Exception as e:
            raise ValueError(
                "Failed to decode outputs for function"
                f" {self.name or self.selector}: {str(e)}"
            )


@dataclass
class ContractABI:
    functions: List[ABIFunction]
    name: Optional[str] = None



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="pyrevm_contract",
    version="0.1.2",
    author="jalbrekt85",
    author_email="jcalbrecht85@gmail.com",
    description="Minimal Brownie like contract wrapper for Pyrevm",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jalbrekt85/pyrevm_contract",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
    install_requires=[
        "pyrevm",
        "eth-abi",
        "pysha3",
    ],

)
